/* Waddle WAD visualizer */

var width = 640;
var height = 400;

var canvas = document.getElementById("main-canvas");
canvas.width = width;
canvas.height = height;

var ctx = canvas.getContext('2d');

function clear() {
    ctx.clearRect(0,0,width,height);
}

var userScale = 1;
var userXOffset = 0;
var userYOffset = 0;

var drawLineDefs = true;
var drawSegs = false;
var drawNodes = false;
var drawGrid = false;
var drawThings = false;

var redraw = true;

var viewMode = 'level';

var currentLevel = '';
var currentLevelFrame = 0;
var currentSprite = '';
var currentSpriteIndex = 0;
var animSprites = [];

var currentTexture = '';

var currentFlat = '';

var currentPatch = '';

var player = null;

function searchNode(nodes, i, x, y) {
    var node = nodes[i];

    console.log('searching node ' + i);
    if (x >= node.lbblx && y >= node.lbbly && x <= node.lbbux && y <= node.lbbuy) {
        if (node.leftNodeOrSSector < 0x8000) {
            return searchNode(nodes, node.leftNodeOrSSector, x, y);
        } else {
            return node.leftNodeOrSSector - 0x8000;
        }
    } else if (x >= node.rbblx && y >= node.rbbly && x <= node.rbbux && y <= node.rbbuy) {
        if (node.rightNodeOrSSector < 0x8000) {
            return searchNode(nodes, node.rightNodeOrSSector, x, y);
        } else {
            return node.rightNodeOrSSector - 0x8000;
        }
    } else {
        return null;
    }
}

function drawNode(nodes, i, depth) {
    if (depth > 200) {
        return;
    }
    // if (depth > currentLevelFrame/20) {
    //     return;
    // }
    
    var node = nodes[i];

    ctx.strokeStyle = 'rgba(0,0,0,0.1)';
    ctx.beginPath();
    ctx.moveTo(node.x, node.y);
    ctx.lineTo(node.x + node.dx, node.y + node.dy);
    ctx.stroke();

    if (false) {
    ctx.strokeStyle = 'rgba(255,0,0,0.25)';
    ctx.strokeRect(node.lbblx, node.lbbly, node.lbbux - node.lbblx, node.lbbuy - node.lbbly);
    ctx.strokeStyle = 'rgba(0,0,255,0.25)';
    ctx.strokeRect(node.rbblx, node.rbbly, node.rbbux - node.rbblx, node.rbbuy - node.rbbly);
    }
    if (node.rightNodeOrSSector < 0x8000) {
        drawNode(nodes, node.rightNodeOrSSector, depth + 1);
    } else {
        // ssector
    }
    if (node.leftNodeOrSSector < 0x8000) {
        drawNode(nodes, node.leftNodeOrSSector, depth + 1);
    } else {
        // ssector
    }
}

function drawLevel(s) {
    if (levels[s] === undefined) {
	return;
    }
    clear();
    var level = levels[s];

    var lines = level.linedefs;
    var vertices = level.vertices;
    var ssectors = level.ssectors;
    var segs = level.segs;
    var minx =  1000000;
    var maxx = -1000000;
    var miny =  1000000;
    var maxy = -1000000;
    for (var i = 0; i < lines.length; i++) {
        var l = lines[i];
        minx = Math.min(minx, Math.min(vertices[l.start].x, vertices[l.end].x));
        maxx = Math.max(maxx, Math.max(vertices[l.start].x, vertices[l.end].x));
        miny = Math.min(miny, Math.min(vertices[l.start].y, vertices[l.end].y));
        maxy = Math.max(maxy, Math.max(vertices[l.start].y, vertices[l.end].y));
    }
    
    var mapWidth = maxx-minx;
    var mapHeight = maxy-miny;

    var scale = (Math.max(mapWidth/width, mapHeight/height))*1.1*userScale;

    ctx.save();
    ctx.translate(userXOffset, userYOffset);
    ctx.scale(1/scale, -1/scale);
    ctx.translate(-minx+((width*scale-mapWidth)/2), -maxy-((height*scale-mapHeight)/2));

    ctx.lineWidth = scale;

    if (drawGrid) {
        var minxDown = Math.floor(minx / 128)*128;
        var maxxUp = Math.ceil(maxx / 128)*128;
        var minyDown = Math.floor(miny/128)*128;
        var maxyUp = Math.ceil(maxy/128)*128
        for (var x = minxDown; x <= maxxUp; x += 128) {
	    ctx.strokeStyle = x == 0 ? 'red' : 'rgba(100,100,100,0.5)';
	    ctx.beginPath();
	    ctx.moveTo(x, minyDown);
	    ctx.lineTo(x, maxyUp);
	    ctx.stroke();
        }
        for (var y = minyDown; y <= maxyUp; y += 128) {
	    ctx.strokeStyle = y == 0 ? 'red' : 'rgba(100,100,100,0.5)';
	    ctx.beginPath();
	    ctx.moveTo(minxDown, y);
	    ctx.lineTo(maxxUp, y);
	    ctx.stroke();
        }
    }

    if (drawLineDefs) {
        ctx.strokeStyle = 'black';

        for (var i = 0; i < lines.length; i++) {
            var l = lines[i];
            ctx.beginPath();
            ctx.moveTo(vertices[l.start].x, vertices[l.start].y);
            ctx.lineTo(vertices[l.end].x, vertices[l.end].y);
            ctx.stroke();
        }
    }

    if (drawSegs) {
        var segColors = ['red', 'blue', 'green', 'gray'];
        var segs = level.segs;
        
        for (var i = 0; i < segs.length; i++) {
            var seg = segs[i];
            ctx.strokeStyle = segColors[i % segColors.length];
            ctx.beginPath();
            ctx.moveTo(vertices[seg.start].x, vertices[seg.start].y);
            ctx.lineTo(vertices[seg.end].x, vertices[seg.end].y);
            ctx.stroke();
        }

        if (false) {
            var ssectors = level.ssectors;
            for (var i = 0; i < ssectors.length; i++) {
                var ssector = ssectors[i];
                ctx.fillStyle = 'rgba(' + (i % 255) + ',' + (i % 255) + ',' + (i % 255) + ',0.5)';
                ctx.beginPath();
                ctx.moveTo(vertices[segs[ssector.segStart].start].x, vertices[segs[ssector.segStart].start].y);
                for (var j = 1; j < ssector.segCount; i++) {
                    ctx.lineTo(vertices[segs[j+ssector.segStart].start].x, vertices[segs[j+ssector.segStart].start].y);
                }
                ctx.fill();
            }
        }
    }

    if (drawNodes) {
        var nodes = level.nodes;

        drawNode(nodes, nodes.length - 1, 0)

        currentLevelFrame += 1;
        if (currentLevelFrame  > 1000) {
            currentLevelFrame = 0;
        }
        redraw = true;
    }

    if (drawThings) {
        var things = level.things;
        for (var i = 0; i < things.length; i++) {
            var t = things[i];
            ctx.beginPath();
            switch (t.type) {
            case 'Player1StartPos':
            case 'Player2StartPos':
            case 'Player3StartPos':
            case 'Player4StartPos':
	        ctx.fillStyle = 'green';
	        break;
            case 'DeathMatchStartPos':
	        ctx.fillStyle = 'rgb(0,200,0)';
	        break;
            case 'Imp':
	        ctx.fillStyle = 'darkgreen';
	        break;
            case 'TeleportLanding':
	        ctx.fillStyle = 'red';
	        break;
            case 'HealthPotion':
	        ctx.fillStyle = 'blue';
	        break;
            case 'SpiritArmor':
	        ctx.fillStyle = 'rgb(100,100,100)';
	        break;
            case 'AmmoClip':
            case 'BoxOfAmmo':
	        ctx.fillStyle = 'rgb(50,50,50)';
	        break;
            default:
	        ctx.fillStyle = 'white';
	        break;
            }
            ctx.arc(t.x, t.y, 15, 0, 2*Math.PI);
            ctx.fill();
        }
    }
    if (player && player.ssector !== null) {
        var ssector = ssectors[player.ssector];
        ctx.strokeStyle = 'red';
        ctx.lineWidth = scale*3;
        ctx.beginPath();
        var seg;
        for (var i = ssector.segStart; i < ssector.segStart + ssector.segCount; i++) {
            seg = segs[i];
            console.log(seg);
            ctx.moveTo(vertices[seg.start].x, vertices[seg.start].y);
            ctx.lineTo(vertices[seg.end].x, vertices[seg.end].y);
        }
        ctx.moveTo(vertices[seg.end].x, vertices[seg.end].y);
        seg = segs[ssector.segStart];
        ctx.lineTo(vertices[seg.start].x, vertices[seg.start].y);
        ctx.stroke();
    }
    ctx.restore();
}

function drawSprite(s) {
    if (sprites[s] === undefined) {
	return;
    }
    clear();

    var scale = 2;

    var sprite = animSprites.length > 0 ? animSprites[Math.round(currentSpriteIndex / 6) % animSprites.length] : sprites[s];

    if (animSprites.length > 0) {
        redraw = true;
    }
    
    var cx = (width / 2) | 0;
    var cy = (height * 2 / 3) | 0;
    ctx.strokeStyle = 'rgba(255,0,0,0.5)';
    ctx.beginPath();
    ctx.moveTo(0,cy);
    ctx.lineTo(width,cy);
    ctx.moveTo(cx,0);
    ctx.lineTo(cx,height);
    ctx.stroke();
    for (var x = 0; x < sprite.width; x++) {
	var y = 0;
	var col = sprite.columns[x];
	for (var p = 0; p < col.length; p++) {
	    var post = col[p];
	    var tx = cx - sprite.leftOffset*scale + x*scale;
	    for (var i = 0; i < post.pixels.length; i++) {
		var ty = cy - sprite.topOffset*scale + (post.top + i)*scale;
		var idx = post.pixels[i];
		var rgb = palettes[0][idx];
		ctx.fillStyle = 'rgb(' + rgb[0] + ',' + rgb[1] + ',' + rgb[2] + ')';
		ctx.fillRect(tx, ty, scale, scale);
	    }
	}
    }
}

function drawPatch(s) {
    if (patches[s] === undefined) {
	return;
    }
    clear();
    var scale = 2;
    var patch = patches[s];
    var cx = (width / 2) | 0;
    var cy = (height * 2 / 3) | 0;
    ctx.strokeStyle = 'rgba(255,0,0,0.5)';
    ctx.beginPath();
    ctx.moveTo(0,cy);
    ctx.lineTo(width,cy);
    ctx.moveTo(cx,0);
    ctx.lineTo(cx,height);
    ctx.stroke();
    for (var x = 0; x < patch.width; x++) {
	var y = 0;
	var col = patch.columns[x];
	for (var p = 0; p < col.length; p++) {
	    var post = col[p];
	    var tx = cx - patch.leftOffset*scale + x*scale;
	    for (var i = 0; i < post.pixels.length; i++) {
		var ty = cy - patch.topOffset*scale + (post.top + i)*scale;
		var idx = post.pixels[i];
		var rgb = palettes[0][idx];
		ctx.fillStyle = 'rgb(' + rgb[0] + ',' + rgb[1] + ',' + rgb[2] + ')';
		ctx.fillRect(tx, ty, scale, scale);
	    }
	}
    }
}

function drawFlat(s) {
    if (flats[s] === undefined) {
	return;
    }
    clear();
    var flat = flats[s];
    var scale = 4;
    for (var i = 0; i < flat.data.length; i++) {
	var tx = (i % 64 | 0) * scale;
	var ty = (i / 64 | 0) * scale;
	var idx = flat.data[i];
	var rgb = palettes[0][idx];
	ctx.fillStyle = 'rgb(' + rgb[0] + ',' + rgb[1] + ',' + rgb[2] + ')';
	ctx.fillRect(tx, ty, scale, scale);
    }
}

function drawTexture(s) {
    if (textures[s] === undefined) {
	return;
    }
    clear();
    var texture = textures[s];
    var scale = 2;
    var minx = 10;
    var maxx = minx + texture.width;
    var miny = 10;
    var maxy = miny + texture.height;
    for (var patchIdx = 0; patchIdx < texture.patches.length; patchIdx++) {
	var patchDesc = texture.patches[patchIdx];
	var patchName = pnames[patchDesc.pname];
	var patch = patches[patchName];

	var cx = patchDesc.xoffset;
	var cy = patchDesc.yoffset;

	for (var x = 0; x < patch.width; x++) {
	    var y = 0;
	    var col = patch.columns[x];
	    for (var p = 0; p < col.length; p++) {
		var post = col[p];
//		var tx = cx - patch.leftOffset + x;
		var tx = cx + x;
		for (var i = 0; i < post.pixels.length; i++) {
//		    var ty = cy - patch.topOffset + (post.top + i);
		    var ty = cy + (post.top + i);
		    var idx = post.pixels[i];
		    var rgb = palettes[0][idx];
		    if (tx >= 0 && tx <= texture.width &&
			ty >= 0 && ty <= texture.height) {
			ctx.fillStyle = 'rgb(' + rgb[0] + ',' + rgb[1] + ',' + rgb[2] + ')';
		    } else {
			ctx.fillStyle = 'rgba(' + rgb[0] + ',' + rgb[1] + ',' + rgb[2] + ',0.1)';
		    }
		    ctx.fillRect((minx+tx)*scale, (miny+ty)*scale, scale, scale);
		}
	    }
	}
    }
    ctx.strokeStyle = 'black';
    ctx.strokeRect(minx*scale,miny*scale,texture.width*scale,texture.height*scale);
}

function frame() {
    requestAnimationFrame(frame);
    currentSpriteIndex += 1;
    if (!redraw) {
        return;
    }
    redraw = false;
    switch (viewMode) {
    case 'level':
        drawLevel(currentLevel);
        break;
    case 'texture':
        drawTexture(currentTexture);
        break;
    case 'patch':
        drawPatch(currentPatch);
        break;
    case 'flat':
        drawFlat(currentFlat);
        break;
    case 'sprite':
        drawSprite(currentSprite);
        break;
    }
}

function chooseLevel(s) {
    viewMode = 'level';
    currentLevel = s;
    redraw = true;
    player = null;
    if (currentLevel in levels) {
        var things = levels[currentLevel].things;
        var nodes = levels[currentLevel].nodes;
        for (var i = 0; i < things.length; i++) {
            if (things[i].type == "Player1StartPos") {
                player = {x: things[i].x,
                          y: things[i].y,
                          angle: things[i].angle*Math.PI/180
                         };
                console.dir(player);
                var ssector = searchNode(nodes, nodes.length - 1, player.x, player.y);
                player.ssector = ssector;
            }
        }
    }
}

var levelSelect = document.getElementById('levels');
for (var s in levels) {
    var o = document.createElement('option');
    if (currentLevel == '') {
        chooseLevel(s);
        o.selected = true;
    }
    o.text = s;
    levelSelect.add(o);
}
levelSelect.addEventListener('change', function(e) {
    chooseLevel(this.options[this.selectedIndex].text);
});

select = document.getElementById('sprites');
for (var s in sprites) {
    var o = document.createElement('option');
    o.text = s;
    select.add(o);
}
select.addEventListener('change', function(e) {
    viewMode = 'sprite';
    currentSprite = this.options[this.selectedIndex].text;
    var animation = currentSprite.substr(0, 4);
    animSprites = [];
    for (var k in sprites) {
        if (k.substr(0,4) == animation) {
            animSprites.push(sprites[k]);
        }
    }
    redraw = true;
});

select = document.getElementById('patches');
for (var s in patches) {
    var o = document.createElement('option');
    o.text = s;
    select.add(o);
}
select.addEventListener('change', function(e) {
    viewMode = 'patch';
    currentPatch = this.options[this.selectedIndex].text;
    redraw = true;
});

select = document.getElementById('flats');
for (var s in flats) {
    var o = document.createElement('option');
    o.text = s;
    select.add(o);
}
select.addEventListener('change', function(e) {
    viewMode = 'flat';
    currentFlat = this.options[this.selectedIndex].text;
    redraw = true;
});

select = document.getElementById('textures');
for (var s in textures) {
    var o = document.createElement('option');
    o.text = s;
    select.add(o);
}
select.addEventListener('change', function(e) {
    viewMode = 'texture';
    currentTexture = this.options[this.selectedIndex].text;
    redraw = true;
});

var gridCheckbox = document.getElementById('drawGrid');
gridCheckbox.addEventListener('change', function(e) {
    drawGrid = !drawGrid;
    redraw = true;
});

var lineDefsCheckbox = document.getElementById('drawLineDefs');
lineDefsCheckbox.addEventListener('change', function(e) {
    drawLineDefs = !drawLineDefs;
    redraw = true;
});

var segsCheckbox = document.getElementById('drawSegs');
segsCheckbox.addEventListener('change', function(e) {
    drawSegs = !drawSegs;
    redraw = true;
});

var nodesCheckbox = document.getElementById('drawNodes');
nodesCheckbox.addEventListener('change', function(e) {
    drawNodes = !drawNodes;
    currentLevelFrame = 0;
    redraw = true;
});

var thingsCheckbox = document.getElementById('drawThings');
thingsCheckbox.addEventListener('change', function(e) {
    drawThings = !drawThings;
    redraw = true;
});

function resizeHandler(e) {
    width = window.innerWidth;
    height = window.innerHeight - 80;
    canvas.width = width;
    canvas.height = height;
}

window.addEventListener('resize', resizeHandler);

resizeHandler();

requestAnimationFrame(frame);
